<?php

/**
 * @file
 * Theme preprocessor functions.
 */
use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Render\Element;

/**
 * Theme preprocessor for 'wallet_wrapper'.
 */
function template_preprocess_mcapi_wallet(&$vars) {
  $vars['canonical'] = $vars['element']['#mcapi_wallet']->toUrl('canonical')->toString();
  $vars['attributes']['class'] = ['wallet', $vars['element']['#view_mode']];
}

/**
 * Theme preprocessor for 'wallet_wrapper'.
 */
function template_preprocess_mcapi_wallet_component(&$vars) {
  $vars['title'] = $vars['element']['#title'];
}

/**
 * Theme preprocessor for all of an entity's wallets.
 */
function template_preprocess_mcapi_wallets(&$vars) {
  $viewbuilder = \Drupal::entityTypeManager()->getViewBuilder('mcapi_wallet');
  // Convert each wallet into a render array.
  foreach ($vars['wallets'] as $key => $wallet) {
    $vars['wallets'][$key] = $viewbuilder->view($wallet, $vars['view_mode']);
  }
}

/**
 * Theme preprocessor for 'balance_bars'.
 */
function template_preprocess_wallet_balance_bars(&$vars) {
  $wallet = $vars['element']['#wallet'];
  $currency = $vars['element']['#currency'];
  $stats = $wallet->getStats($currency->id);
  $id = "given-received-" . $currency->id . '-' . $wallet->id();
  $max = _mcapi_get_axis_max(
    $currency->format(max($stats->gross_in, $stats->gross_out), CurrencyInterface::DISPLAY_PLAIN)
  );
  $vars += [
    'label' => $currency->label(),
    'id' => $id,
    'functionname' => str_replace('-', '_', $id),
    'incoming' => $currency->format($stats->gross_in, CurrencyInterface::DISPLAY_PLAIN),
    'outgoing' => $currency->format($stats->gross_out, CurrencyInterface::DISPLAY_PLAIN),
    'show_in' => $currency->format($stats->gross_in),
    'show_out' => $currency->format($stats->gross_out),
    'given' => t('Given'),
    'received' => t('Received'),
    'max' => $max,
    'vaxislabels' => [
      [
        'value' => 0,
        'label' => strip_tags($currency->format(0, CurrencyInterface::DISPLAY_NORMAL)),
      ],
      [
        'value' => $max / 2,
        'label' => strip_tags($currency->format(intval($max / 2), CurrencyInterface::DISPLAY_NORMAL)),
      ],
      [
        'value' => $max,
        'label' => strip_tags($currency->format($max, CurrencyInterface::DISPLAY_NORMAL)),
      ],
    ],
  ];
}

/**
 * Theme preprocessor for 'wallet_stats'.
 */
function template_preprocess_wallet_stats(&$vars) {
  $unused = \Drupal::config('mcapi.settings')->get('render_unused');
  foreach ($vars['element']['#wallet']->getAllStats($unused) as $curr_id => $data) {
    $currency = Currency::load($curr_id);
    $vars['currencies'][$curr_id]['label'] = $currency->label();
    $vars['currencies'][$curr_id]['gross_in'] = $currency->format($data->gross_in, CurrencyInterface::DISPLAY_NORMAL);
    $vars['currencies'][$curr_id]['gross_out'] = $currency->format($data->gross_out, CurrencyInterface::DISPLAY_NORMAL);
    $vars['currencies'][$curr_id]['volume'] = $currency->format($data->volume, CurrencyInterface::DISPLAY_NORMAL);
    $vars['currencies'][$curr_id]['balance'] = $currency->format($data->balance, CurrencyInterface::DISPLAY_NORMAL);
    $vars['currencies'][$curr_id]['partners'] = $data->partners;
    $vars['currencies'][$curr_id]['trades'] = $data->trades;
  }
}


/**
 * Helper.
 *
 * @param int $val
 *   The max value to represent in the chart.
 */
function _mcapi_get_axis_max($val) {
  $max = 10 ** ceil(log($val, 10));
  while ($max/2 > $val) {
    $max = $max / 2;
  }
  return $max;
}

/**
 * Theme preprocessor for 'mcapi_transaction'.
 *
 * Can't work out why, but we need to render all the transaction properties in
 * this function because twig seems to skip them
 */
function template_preprocess_mcapi_transaction(&$vars, $hook, $info) {
  $elements = $vars['elements'];
  foreach (Element::children($vars['elements']) as $key) {
    $vars[$key] = render($vars['elements'][$key]);
  }
  if (isset($elements['#watermark'])) {
    // to add a watermark to the render array, add #watermark => blah
    $vars['watermark'] = Markup::create('<div class = "watermark">'.$elements['#watermark'].'</div>');
  }
  if ($elements['#mcapi_transaction']->children) {
    $children = \Drupal::entityTypeManager()
      ->getViewBuilder('mcapi_transaction')
      ->viewMultiple($elements['#mcapi_transaction']->children, 'sentence');
    $vars['children'] = \Drupal::service('renderer')->render($children);
  }
  $vars['links'] = $elements['#links'];
}

/**
 * Theme preprocessor for 'mcapi_timeline'.
 *
 * Adds a column to the data with the formatted currency value.
 */
function template_preprocess_mcapi_timeline(&$vars) {
  $vars['id'] = preg_replace('/[^a-z0-9_]+/', '_', strtolower($vars['title']));
  foreach ($vars['points'] as $timestamp => $val) {
    $vars['daterows'][] = [
    // Format for javascript dates.
      date('Y,m,d', $timestamp),
      $val,
    ];
  }
  // TEMP until gcharts works better. Force some formatted values on the axis.
  if ($currency = $vars['currency']) {
    list($min, $middle, $max) = gchart_axes($currency, $vars['points']);
    $vars['vaxislabels'] = [
      [
        'value' => $min,
        'label' => $currency->format($min, CurrencyInterface::DISPLAY_NORMAL),
      ],
      [
        'value' => $middle,
        'label' => $currency->format($middle, CurrencyInterface::DISPLAY_NORMAL),
      ],
      [
        'value' => $max,
        'label' => $currency->format($max, CurrencyInterface::DISPLAY_NORMAL),
      ],
    ];
  }
  // @todo is there a way to have html labels in google charts?
  if ($vars['vaxislabels']) {
    foreach ($vars['vaxislabels'] as &$tick) {
      $tick['label'] = strip_tags($tick['label']);
    }
  }
}

/**
 * Calculate some good axis labels, using the min & max balance extents.
 *
 * @param array $vals
 *   The $points history of raw balances keyed by unixtimes.
 *
 * @return array
 *   Values keyed by min, 0 & max.
 */
function gchart_axes(CurrencyInterface $currency, $vals) {
  $max = $currency->format($vals ? max($vals) : 0, CurrencyInterface::DISPLAY_PLAIN);
  $min = $currency->format($vals ? min($vals) : 0, CurrencyInterface::DISPLAY_PLAIN);
  if ($min >= 0) {
    $max = _mcapi_get_axis_max($max);
    return [0, $max / 2, $max];
  }
  elseif ($max <= 0) {
    $min = -_mcapi_get_axis_max(abs($min));
    return [$min, $min / 2, 0];
  }
  else {
    return [-_mcapi_get_axis_max(abs($min)), 0, _mcapi_get_axis_max($max)];
  }
}

function template_preprocess_mcapi_currency(&$variables) {
  $currency = $variables['elements']['#mcapi_currency'];
  $stats = \Drupal::entityTypeManager()
    ->getStorage('mcapi_transaction')
    ->ledgerStateQuery($currency->id(), [])
    ->execute()
    ->fetch();
  $variables['name'] = $currency->label();
  $variables['volume'] = $currency->format($stats->volume, CurrencyInterface::DISPLAY_NORMAL, TRUE);
  $variables['trades'] = $stats->trades;
  $variables['wallets'] = $stats->partners;
  $variables['url'] = $currency->toUrl();
}

function template_preprocess_viz_balance(&$vars) {
  $currency = Currency::load($vars['curr_id']);
  if ($vars['gross_in'] > $vars['gross_out']) {
    $vars['side'] = 'positive';
  }
  elseif ($vars['gross_in'] < $vars['gross_out']) {
    $vars['side'] = 'negative';
  }
  else {
    $vars['side'] = 'balanced';
  }
  $vars['income'] = $currency->format($vars['gross_in']);
  $vars['expenditure'] = $currency->format($vars['gross_out']);
  $vars['balance'] = $currency->format($vars['gross_in'] - $vars['gross_out']);
}