<?php

namespace Drupal\mcapi\Entity\Storage;

use Drupal\mcapi\Entity\Transaction;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Cache\Cache;

/**
 * Storage controller for Transaction entity.
 */
class TransactionStorage extends TransactionIndexStorage {

  /**
   * {@inheritdoc}
   */
  public function loadBySerial($serial, $skip_access = FALSE) : Transaction {
    $transactions = $this->loadByProperties(['serial' => $serial]);
    return reset($transactions);
  }

  public function loadMultiple(array $xids = NULL) {
    // This works with xids, because views needs it.
    return parent::loadMultiple($xids);
  }

  /**
   * {@inheritdoc}
   *
   * Because the transaction entity is keyed by serial number not xid,
   * and because it contains child entities,
   * We need to overwrite the whole save function
   * and by the time we call the parent, we pass it individual transaction
   * entities having called $transaction->flatten.
   */
  public function save(EntityInterface $transaction) {
    //we can't start this process while another user might still be saving a
    //transaction because of the way the serial number is determined
    // Determine the serial number.
    if (empty($transaction->serial->value)) {
      $transaction->serial->value = $this->getSerial();
    }
    // Note that flatten() clones the transactions.
    foreach ($transaction->flatten(FALSE) as $entity) {
      $entity->serial->value = $transaction->serial->value;
      // Entity parent is 0 for the first entity and then the xid of the first
      // for all children saved consequently.
      $entity->parent->value = $parent??0;
      $return = parent::save($entity);
      if (!isset($parent)) {
        $parent = $entity->id();
      }
    }
    $transaction->enforceIsNew(FALSE);
    return $return;
  }

  private function getSerial() : int {
    $last = \Drupal::state()->get('mcapi.last_serial');
    if (!$last) { //TEMP
      $last = $this->database->query(
        "SELECT MAX([serial]) FROM {mcapi_transaction}"
      )->fetchField();
    }
    $next = $last+1;
    \Drupal::state()->set('mcapi.last_serial', $next);
    return $next;
  }

  /**
   * {@inheritdoc}
   */
  public function doSave($id, EntityInterface $entity) {
    $record = $this->mapToStorageRecord($entity);
    $record->changed = \Drupal::time()->getRequestTime();
    $return = parent::doSave($entity->xid->value, $entity);
    // The entity is no longer new.
    // because we were working on a clone.
    $entity->enforceIsNew(FALSE);

    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function resetCache(array $ids = NULL) {
    parent::resetCache($ids);
  }

  /**
   * {@inheritdoc}
   */
  protected function doDelete($entities) {
    foreach ($entities as $transaction) {
      $this->invalidateWalletCache($transaction);
    }
    parent::doDelete($entities);
  }

  /**
   * @param EntityInterface $entity
   * @param bool $update
   */
  protected function doPostSave(EntityInterface $entity, $update) {
    $this->invalidateWalletCache($entity);
    parent::doPostSave($entity, $update);
  }

  /**
   * @param Transaction $transaction
   */
  private function invalidateWalletCache(Transaction $transaction) {
    $payer_tags = $payee_tags = [];
    foreach ($transaction->flatten() as $t) {
      // in cases of migration rollbacks the wallet may not be present
      if ($payer_wallet = $t->payer->entity) {
        $payer_tags = $payer_wallet->getCacheTagsToInvalidate();
      }
      if ($payee_wallet = $t->payee->entity) {
        $payee_tags = $payee_wallet->getCacheTagsToInvalidate();
      }
    }
    Cache::invalidateTags(array_merge($payer_tags, $payee_tags));
  }

}
