<?php

namespace Drupal\mcapi\Entity;

use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\Type;
use Drupal\mcapi\Entity\State;
use Drupal\mcapi\Event\TransactionAssembleEvent;
use Drupal\mcapi\Event\McapiEvents;
use Drupal\user\UserInterface;
use Drupal\user\EntityOwnerTrait;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityChangedTrait;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Entity\EntityTypeInterface;

/**
 * Defines the Transaction entity.
 *
 * Transactions are grouped and handled by serial number but the unique database
 * key is the xid.
 *
 * @ContentEntityType(
 *   id = "mcapi_transaction",
 *   label = @Translation("Transaction"),
 *   label_collection = @Translation("Transactions"),
 *   handlers = {
 *     "storage" = "Drupal\mcapi\Entity\Storage\TransactionStorage",
 *     "storage_schema" = "Drupal\mcapi\Entity\Storage\TransactionStorageSchema",
 *     "view_builder" = "Drupal\mcapi\Entity\ViewBuilder\TransactionViewBuilder",
 *     "list_builder" = "\Drupal\mcapi\Entity\ListBuilder\TransactionListBuilder",
 *     "access" = "Drupal\mcapi\Entity\Access\TransactionAccessControlHandler",
 *     "form" = {
 *       "default" = "Drupal\mcapi\Form\TransactionForm",
 *       "bill" = "Drupal\mcapi\Form\TransactionForm",
 *       "credit" = "Drupal\mcapi\Form\TransactionForm",
 *       "edit" = "Drupal\mcapi\Form\TransactionEditForm",
 *     },
 *     "views_data" = "Drupal\mcapi\Entity\Views\TransactionViewsData",
 *     "route_provider" = {
 *       "html" = "Drupal\mcapi\Entity\RouteProvider\TransactionRouteProvider",
 *     },
 *   },
 *   admin_permission = "configure mcapi",
 *   base_table = "mcapi_transaction",
 *   entity_keys = {
 *     "id" = "xid",
 *     "uuid" = "uuid",
 *     "name" = "description",
 *     "owner" = "creator"
 *   },
 *   field_ui_base_route = "entity.mcapi_transaction.collection",
 *   translatable = FALSE,
 *   links = {
 *     "canonical" = "/transaction/{mcapi_transaction}",
 *     "add-form" = "/transaction/add",
 *     "edit-form" = "/transaction/{mcapi_transaction}/edit",
 *     "collection" = "/admin/accounting/transactions"
 *   },
 *   constraints = {
 *     "DifferentWallets" = {},
 *     "SameState" = {},
 *   }
 * )
 */
class Transaction extends ContentEntityBase implements TransactionInterface {

  use MessengerTrait;
  use EntityChangedTrait;
  use EntityOwnerTrait;

  private $moduleHandler;
  protected $eventDispatcher;

  /**
   * Constructor.
   *
   * @note There seems to be no way to inject anything here.
   *
   * @see SqlContentEntityStorage::mapFromStorageRecords
   */
  public function __construct(array $values, $entity_type, $bundle = FALSE, $translations = array()) {
    parent::__construct($values, $entity_type, $bundle, $translations);
    $this->moduleHandler = \Drupal::moduleHandler();
    $this->eventDispatcher = \Drupal::service('event_dispatcher');
  }

  /**
   * {@inheritdoc}
   */
  public function label($langcode = NULL) {
    return t("Transaction #@serial", ['@serial' => $this->serial->value]);
  }

  /**
   * {@inheritdoc}
   * Trying to see if this is ever called with the xid.
   * Any problems just delete this.
   */
  public static function load($serial) {
    return static::loadBySerial($serial);
  }

  /**
   * {@inheritdoc}
   */
  public static function loadBySerial(int $serial) : Transaction {
    return \Drupal::entityTypeManager()
      ->getStorage('mcapi_transaction')
      ->loadBySerial($serial);
  }

  /**
   * {@inheritdoc}
   */
  protected function urlRouteParameters($rel) {
    return [
      'mcapi_transaction' => $this->serial->value,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function assemble() {
    if ($this->parent->value) {
      // This should never happen
      // Assembly of child transactions could cause infinite recursion.
      return;
    }
    // Pass the transaction round to add children, but prevent the transaction
    // from being modified by sending a clone.
    $clone = clone($this);
    // This hook prevents altering the original transaction but only allows
    // adding of children.
    $event = new TransactionAssembleEvent($clone, ['mcapi_transaction' => $clone]);
    $this->eventDispatcher->dispatch(McapiEvents::ASSEMBLE, $event);
    $this->children = $clone->children;
    // Allow any module to alter the transaction & children, before validation.
    $this->moduleHandler->alter('mcapi_transaction_assemble', $this);

    foreach ($this->children as $child) {
      // Temp value because parent has no xid yet.
      $child->parent->value = 1;
    }
  }

  /**
   * Validate a transaction (including children).
   */
  public function validate() {
    $this->assemble();
    $violationList = parent::validate();
    if ($this->parent->value == 0 and isset($this->children)) {
      foreach ($this->children as $entity) {
        if ($child_violations = $entity->validate()) {
          $violationList->addAll($child_violations);
        }
      }
    }
    return $violationList;
  }

  /**
   * {@inheritdoc}
   */
  public static function preCreate(EntityStorageInterface $storage, array &$values) {
    $type_name = isset($values['type']) ? $values['type'] : 'default';
    if ($type = Type::load($type_name)) {
      // doesn't apply to imported transactions or child transactions.
      if (empty($values['serial']) and empty($values['state'])) {
        $values['state'] = $type->start_state;
      }
    }
    else {
      trigger_error("Replaced bad transaction type '$type_name' with 'default'", E_USER_ERROR);
    }
    if (empty($values['worth']) and isset($values['quantity'])) {
      if ($curr = $values['currency'] ?? $values['currcode'] ?? $values['curr_id']) {
        $values['worth'][] = [
          'curr_id' => $curr,
          'value' => $values['quantity']
        ];
        unset($values['currency'], $values['currcode'], $values['curr_id']);
      }
      else {
        throw new \Exception('Cannot derive worth field from given fields');
      }
    }
    $values += [
      'type' => 'default',
      // Uid of 0 means drush must have created it.
      'creator' => static::getDefaultEntityOwner(),
      'state' => 'done',
      'children' => []
    ];

    // There isn't a hook these can be moved into.
    if (isset($values['payer']) and is_array($values['payer'])) {
      $values['type'] = 'mass';
      $payers = $values['payer'];
      $values['payer']= array_shift($payers);
      foreach($payers as $payer) {
        $values['children'][] = static::create(['payer' => $payer] + $values);
      }
    }
    elseif (isset($values['payee']) and is_array($values['payee'])) {
      $values['type'] = 'mass';
      $payees = $values['payee'];
      $values['payee'] = array_shift($payees);
      foreach($payees as $payee) {
        $values['children'][] = static::create(['payee' => $payee] + $values);
      }
    }
  }

  function setType(string $type_id) {
    if ($this->isNew()) {
      $this->type->target_id = $type_id;
      $this->state->target_id = Type::load($type_id)->start_state;
    }
    else {
      throw new \Exception('Too late to set transaction type.');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTagsToInvalidate() {
    $tags = parent::getCacheTagsToInvalidate();
    // Invalidate tags of wallets in child transactions.
    foreach ($this->flatten(FALSE) as $transaction) {
      foreach (['payer', 'payee'] as $actor) {
        if ($wallet = $transaction->{$actor}->entity) {
          $tags = array_merge_recursive($tags, $wallet->getCacheTagsToInvalidate());
        }
      }
    }
    // We should only be clearing the currency displays, but I haven't worked out the cache tags for that.
    foreach ($this->worth->currencies(TRUE) as $currency) {
      $tags = array_merge($tags, $currency->getCacheTagsToInvalidate());
    }
    return $tags;
  }

  /**
   * Take a load a of flat transaction entities keyed by xid and put the in a
   * parent/child relationships, keyed by xid (expected by views).
   */
  public static function postLoad(EntityStorageInterface $storage_controller, array &$entities) {
    ksort($entities);
    foreach ($entities as $key => $entity) {
      if ($entity->parent->value == 0) {
        // This is a parent
        $keyed_by_serial[$entity->xid->value] = $entity;
      }
      elseif (isset($keyed_by_serial[$entity->parent->value])) {
        // This is a child with a parent.
        $keyed_by_serial[$entity->parent->value]->children[] = $entity;
      }
      else {
        //This is a child by itself - load it as if it were standalone.
        $keyed_by_serial[$entity->xid->value] = $entity;
      }
    }
    $entities = $keyed_by_serial;
  }

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type);
    $fields += static::ownerBaseFieldDefinitions($entity_type);
    $fields['description'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Description'))
      ->setDescription(t('The good or service.'))
      ->setSettings(['max_length' => 255])
      ->setRequired(FALSE)
      ->setDefaultValue('')
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('form', ['type' => 'string_textfield', 'weight' => 3])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', ['type' => 'string', 'weight' => 3]);

    $fields['serial'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Serial number'))
      ->setDescription(t('Grouping of related transactions.'))
      ->setReadOnly(TRUE)
      ->setRequired(FALSE);

    $fields['parent'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Parent xid'))
      ->setDescription(t('Parent transaction that created this transaction.'))
      ->setDefaultValue(0)
      ->setSetting('target_type', 'mcapi_transaction')
      ->setReadOnly(TRUE);

    $fields['worth'] = BaseFieldDefinition::create('worth')
      ->setLabel(t('Worth'))
      ->setDescription(t('Value of the transaction (one or more currencies)'))
      ->setCardinality(-1)
      ->setRequired(TRUE)
      ->addPropertyConstraints('curr_id', [
        'AllowedValues' => ['callback' =>  ['Drupal\mcapi\Entity\Transaction', 'getAllCurrencyIds']],
      ])
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', ['weight' => 4]);

    $fields[PAYER_FIELDNAME] = BaseFieldDefinition::create('wallet_reference')
      ->setLabel(t('Payer'))
      ->setDescription(t('The giving wallet'))
      ->setReadOnly(TRUE)
      ->setRequired(TRUE)
      ->setCardinality(1)
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('form', ['type' => 'wallet_reference_autocomplete', 'weight' => 1])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', ['type' => 'entity_reference_label', 'weight' => 1]);

    $fields[PAYEE_FIELDNAME] = BaseFieldDefinition::create('wallet_reference')
      ->setLabel(t('Payee'))
      ->setDescription(t('The receiving wallet'))
      ->setReadOnly(TRUE)
      ->setRequired(TRUE)
      ->setCardinality(1)
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('form', ['type' => 'wallet_reference_autocomplete', 'weight' => 2])
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('view', ['type' => 'entity_reference_label', 'weight' => 2]);

    $fields['creator'] // Alterations to ownerBaseFieldDefinition
      ->setLabel(t('Creator'))
      ->setDescription(t('The user who created the transaction'))
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions(
        'form',
        ['type' => 'hidden', 'weight' => 10]
      )
      ->setReadOnly(TRUE)
      ->setRequired(TRUE)
      ->setRevisionable(FALSE);

    $fields['created'] = BaseFieldDefinition::create('created')
      ->setLabel(t('Created on'))
      ->setDescription(t('The time the transaction was created.'))
      ->setRevisionable(FALSE)
      ->setRequired(TRUE)
      ->setDisplayOptions(
        'form',
        ['type' => 'hidden', 'weight' => 10]
      )
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE);

    $fields['changed'] = BaseFieldDefinition::create('changed')
      ->setLabel(t('Changed'))
      ->setDescription(t('The time that the transaction was last saved.'))
      ->setDisplayConfigurable('view', TRUE)
      ->setInitialValue('created')
      ->setRevisionable(FALSE)
      ->setRequired(TRUE)
      ->setTranslatable(FALSE);

    $fields['type'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Type'))
      ->setDescription(t('The type/workflow path of the transaction'))
      ->setSetting('target_type', 'mcapi_type')
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayOptions('form', ['region' => 'hidden'])
      ->addPropertyConstraints('target_id', [
        'AllowedValues' => ['callback' =>  ['Drupal\mcapi\Entity\Transaction', 'getAllowedTypes']],
      ])
      ->setReadOnly(TRUE)
      ->setRequired(TRUE);

    $fields['state'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('State'))
      ->setDescription(t('Completed, pending, disputed, etc.'))
      ->setSetting('target_type', 'mcapi_state')
      ->setDisplayConfigurable('view', TRUE)
      ->setDisplayOptions('form', ['region' => 'hidden'])
      ->addPropertyConstraints('target_id', [
        'AllowedValues' => ['callback' => ['Drupal\mcapi\Entity\Transaction', 'getAllowedStates']],
      ])
      ->setReadOnly(FALSE)
      ->setRequired(TRUE);
    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function flatten($clone = TRUE) {
    if ($clone) {
      $transaction = clone($this);
    }
    else {
      $transaction = $this;
    }
    $flatarray = [$transaction];
    if (!empty($transaction->children)) {
      foreach ($transaction->children as $child) {
        // These transactions haven't necessarily been saved, so don't have an id
        $flatarray[] = $child;
      }
    }
    return $flatarray;
  }

  /**
   * {@inheritdoc}
   */
  public function getOwner() {
    return $this->creator->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getOwnerId() {
    return $this->creator->target_id;
  }

  /**
   * {@inheritdoc}
   */
  public function setOwner(UserInterface $account) {
    $this->creator = $account;
  }

  /**
   * {@inheritdoc}
   */
  public function setOwnerId($uid) {
    $this->creator->targetId = $uid;
  }

  /**
   * {@inheritdoc}
   */
  public function getCreatedTime() {
    return $this->get('created')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function toArray() {
    $array = parent::toArray();
    if (isset($array['children'])) {
      foreach (@$array['children'] as &$transaction) {
        $transaction = $transaction->toArray();
      }
    }
    return $array;
  }


  /**
   * {@inheritdoc}
   */
  static function getAllowedStates() :array {
    return array_keys(State::LoadMultiple());
  }

  /**
   * {@inheritdoc}
   */
  static function getAllowedTypes() {
    return array_keys(Type::LoadMultiple());
  }

  /**
   * {@inheritdoc}
   */
  static function getAllCurrencyIds() {
    return array_keys(Currency::LoadMultiple());
  }
  /**
   * {@inheritdoc}
   * @deprecated but removing requires reinstalling the field definition and content as in mcapi_update_8998
   */
  static function getAllowedCurrencies() {
    return static::getAllCurrencyIds();
  }

  /**
   * {@inheritdoc}
   * @todo consider renaming the id key to 'id' and remove this.
   */
  public function id() {
    return $this->xid->value ?? NULL;
  }

  /**
   * {@inheritdoc}
   *
   * Workaround for when the content_translation module tries to link to the
   * original language before the entity is saved and an id is available on transaction/0/save
   * I don't know why the entity definition 'translatable = FALSE' doesn't apply
   */
  public function hasLinkTemplate($rel) {
    if ($this->isNew()) return FALSE;
    return parent::hasLinkTemplate($rel);
  }

}

