<?php

namespace Drupal\mcapi\Form;

use Drupal\mcapi\Event\McapiEvents;
use Drupal\mcapi\Event\TransactionSaveEvents;
use Drupal\mcapi\TransactionOperations;
use Drupal\mcapi\Entity\Transaction;
use Drupal\Core\Url;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Render\Renderer;
use Drupal\Core\Render\Markup;
use Drupal\Core\Utility\Token;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * A form for confirming all operations on a transaction.
 */
class OperationForm extends ContentEntityForm {

  /**
   * @var Drupal\system\Entity\Action;
   */
  private $plugin;

  /**
   *
   * @var array
   */
  private $config;
  private $eventDispatcher;
  private $renderer;
  private $destinationPath;
  protected $transactionViewBuilder;

  /**
   *
   * @param EntityRepositoryInterface $entity_repository
   * @param EntityTypeBundleInfoInterface $entity_type_bundle_info
   * @param TimeInterface $time
   * @param RequestStack $request_stack
   * @param ContainerAwareEventDispatcher $event_dispatcher
   * @param Renderer $renderer
   * @param Token $token
   */
  public function __construct(EntityRepositoryInterface $entity_repository, EntityTypeBundleInfoInterface $entity_type_bundle_info = NULL, TimeInterface $time = NULL, EntityTypeManager $entity_type_manager, RequestStack $request_stack, ContainerAwareEventDispatcher $event_dispatcher, Renderer $renderer, Token $token) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
    $this->transactionViewBuilder = $entity_type_manager->getViewBuilder('mcapi_transaction');
    $this->eventDispatcher = $event_dispatcher;
    $this->renderer = $renderer;
    $this->token = $token;
    $query = $request_stack->getCurrentRequest()->query;
    if ($query->has('destination')) {
      $this->destinationPath = $query->get('destination');
    }
  }

  protected function init(FormStateInterface $form_state) {
    parent::init($form_state);
    // These can't be initiated in construct because operation hasn't been set by then.
    $this->plugin = TransactionOperations::loadOperation($this->getOperation());
    $this->config = $this->plugin->getPlugin()->getConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('entity_type.manager'),
      $container->get('request_stack'),
      $container->get('event_dispatcher'),
      $container->get('renderer'),
      $container->get('token')
    );
  }

  /**
   * Makes the protected transaction $entity accessible via form_alter hooks.
   *
   * @return Transaction
   */
  function getTransaction() : Transaction {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'mcapi_' . $this->getOperation() . '_form';
  }


  public function setOperation($operation) {
    // If NULL is passed, do not overwrite the operation.
    if ($operation) {
      $this->operation = $operation;
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    // store the children because on submission the entity will be rebuilt but without validation.
    $form_state->set('children', $this->entity->children);

    $form['#title'] = $this->config['page_title'];
    if ($this->config['format'] == 'twig') {
      $existing_transaction = [
        '#type' => 'inline_template',
        '#attached' => [
          'library' => ['mcapi/mcapi.transaction']
        ],
        '#template' => $this->config['twig'],
        '#context' => ['mcapi_transaction' => $this->entity],
      ];
    }
    else {
      $existing_transaction = $this->transactionViewBuilder->view($this->entity, $this->config['format']);
    }

    $form['transaction'] = $existing_transaction;
    $form['transaction']['#weight'] = -1;

    // Not sure what this is for
    $form['confirm'] = ['#type' => 'hidden', '#value' => 1];

    // By default, render the form using theme_confirm_form().
    if (!isset($form['#theme'])) {
      $form['#theme'] = 'confirm_form';
    }

    $form['#attributes']['class'][] = 'transaction-operation-form';
    $form['actions']['submit']['#value'] = $this->config['button'];
    return $form;
  }

  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Override the default validation implementation as it is not necessary
    // nor possible to validate an entity in a confirmation form.
    return $this->entity = $this->buildEntity($form, $form_state);
  }


  public function buildEntity(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
    $entity = parent::buildEntity($form, $form_state);
    $entity->children = $form_state->get('children');
    return $entity;
  }
  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // Use the entity as is, don't rebuild it
    $transaction = $this->entity;
    $transaction->setValidationRequired(FALSE);
    $this->updateChangedTime($transaction);
    try {
      // The op might have injected values into the form, so it needs to be able
      // to access them.
      $this->plugin->execute([$transaction]);
      if ($this->getOperation() == 'delete') {
        if (empty($this->destinationPath)) {
          // Front page.
          $this->destinationPath = '/';
        }
      }
      $context = [
        'values' => $form_state->getValues(),
        'old_state' => $transaction->state->value,
        'action' => $this->getOperation(),
      ];
      $events = new TransactionSaveEvents($transaction, $context);
      $this->eventDispatcher->dispatch(McapiEvents::ACTION, $events);
      if ($m = $this->config['message']) {
        $this->messenger()->addMessage($m, 'status');
      }
    }
    catch (\Exception $e) {
      $this->messenger()->addError($this->t(
        "Error performing @action action: @error",
        [
          '@action' => $this->config['title'],
          '@error' => $e->getMessage(),
        ]
      ));
      return;
    }

    if ($this->destinationPath) {
      $redirect_url = Url::fromUri('base:' . $this->destinationPath);
    }
    elseif ($transaction->id()) {
      // destructive testing showed that its possible to not have saved at this point
      $redirect_url = $transaction->toUrl();
      $form_state->setRedirectUrl($redirect_url);
    }
  }

  /**
   * {@inheritdoc}
   *
   * @todo can't see how to make a back link using javascript
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = [
      'submit' => [
        '#type' => 'submit',
        '#value' => $this->t('Confirm'),
        '#submit' => ['::submitForm'],
      ]
    ];

    if ($text = $this->config['cancel_link']) {
      // Prepare cancel link.
      $query = $this->getRequest()->query;
      $cancel_url = NULL;
      // Borrowed from ConfirmFormHelper::buildCancelLink
      // If a destination is specified, that serves as the cancel link.
      if ($query->has('destination')) {
        $options = UrlHelper::parse($query->get('destination'));
        // @todo Revisit this in https://www.drupal.org/node/2418219.
        try {
          $cancel_url = Url::fromUserInput('/' . ltrim($options['path'], '/'), $options);
        }
        catch (\InvalidArgumentException $e) {
          // Suppress the exception and fall back to the form's cancel url.
        }
      }
      if (!$cancel_url) {
        if ($this->plugin->getPlugin()->getPluginId() == 'mcapi_transaction.save_action') {
          $cancel_url = new Url('user.page');
        }
        else {
          $cancel_url = $this->entity->toUrl();
        }
      }
      // Wouldn't be surprised if there was a more api-friendly way to do this.
      $actions['cancel'] = [
        '#markup' => Markup::create('<a onclick = "window.history.go(-1)" style="cursor:pointer">'.$text.'</a>')
      ];
    }

    return $actions;
  }


}
