<?php

namespace Drupal\mcapi\Plugin\Block;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Block\BlockBase;
use Drupal\mcapi\Entity\Currency;
use Drupal\Core\Render\Markup;
use Drupal\Core\Database\Driver\mysql\Connection;
/**
 * Displays stats indicating the overall activity on the site.
 *
 * @Block(
 *   id = "mcapi_system_stats",
 *   admin_label = @Translation("System-wide trading stats"),
 *   category = @Translation("Community Accounting")
 * )
 */
class SystemStats extends BlockBase implements ContainerFactoryPluginInterface {

  /**
   * @var Drupal\mysql\Driver\Database\mysql\Connection
   */
  private $database;

  function __construct($configuration, $plugin_id, $plugin_definition, Connection $database) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->database = $database;
  }

  public static function create($container, array $configuration, $plugin_id, $plugin_definition) {
    return new static (
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('database')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function label() {
    if (!empty($this->configuration['label'])) {
      return $this->configuration['label'];
    }
    return t('@period trading stats', ['@period' => $this->configuration['period']]);
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $conf = parent::defaultConfiguration();
    return $conf += ['period' => '1 year'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);
    $form['period'] = [
      '#title' => t('Time since (relative to now)'),
      '#description' => t('E.g. 1 year, 6 months, 30 days. See %link', ['%link' => 'https://www.php.net/manual/en/datetime.formats.relative.php']),
      '#type' => 'textfield',
      '#size' => 10,
      '#default_value' => $this->configuration['period'],
      '#weight' => 5
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function build() {
    $currencies = Currency::loadMultiple();
    $currency = reset($currencies);

    $query = $this->database->select('mcapi_transactions_index', 't');
    $query->addExpression('COUNT(t.serial)', 'trades');
    $query->addExpression('COUNT(DISTINCT t.partner_id)', 'traders');
    $query->addExpression('sum(t.volume)', 'volume');
    $result = $query->condition('t.outgoing', 0)
      ->condition('t.curr_id', $currency->id)
      ->condition('created', strtotime('- '.$this->configuration['period']), '>')
      ->execute()
      ->fetchAssoc();
    $build['#title'] = $this->t(':currency stats', [':currency' => $currency->label()]);
    $build['trades'] = $this->littleDiv($this->t('Trades'), $result['trades']);
    $build['traders'] = $this->littleDiv($this->t('Traders'), $result['traders']);
    $build['volume'] = $this->littleDiv($this->t('Volume'), $currency->format(intval($result['volume'])));
    return $build;
  }

  /**
   * {@inheritdoc}
   */
  function littleDiv(string $title, string $value) {
    static $w = 0;
    $markup = '<strong>'.$title.':</strong> ';
    $markup .= $value;
    return [
      '#type' => 'markup',
      '#prefix' => '<div>',
      '#suffix' => '</div>',
      '#weight' => $w++,
      '#markup' => Markup::create($markup)
    ];
  }


}
