<?php

namespace Drupal\mcapi;

use Drupal\mcapi\Entity\TransactionInterface;
use Drupal\system\Entity\Action;
use Drupal\Core\Url;

/**
 * Handle the transaction entity's 'operations' which are in fact system actions
 * with extra config. Most operations lead to a highly configured confirm form,
 * which can be reached though ajax, a modal box or in a new page.
 * View is special because it has no confirm form and indeed no action, but
 * shares access control and page formatting with the other operations. Save is
 * also special, being hidden from the UI.
 *
 * @todo make this a TransactionActionManager service utilising a Collection.
 */
class TransactionOperations {

  /**
   * Get the operations for the given transaction
   *
   * @param TransactionInterface $transaction
   *
   * @return array
   *   A renderable array
   *
   * @todo cache these by user and transaction
   */
  public static function get(TransactionInterface $transaction, $include_view) {
    $transactionAccessController = \Drupal::entityTypeManager()->getAccessControlHandler('mcapi_transaction');
    $operations = [];
    $actions = static::loadAllActions();
    if (!$include_view) {
      unset($actions[array_search('view', $actions)]);
    }
    foreach ($actions as $name) {
      if ($transactionAccessController->access($transaction, $name)) {
        if ($name == 'view' or $name == 'save') {
          $route_name = 'entity.mcapi_transaction.canonical';
        }
        else {
          $route_name = 'entity.mcapi_transaction.'.$name;
        }
        $action_plugin = static::loadOperation($name)->getPlugin();
        $operations[$name] = [
          'title' => $action_plugin->getConfiguration()['title'],
          'url' => Url::fromRoute($route_name, ['mcapi_transaction' => $transaction->serial->value]),
          'weight' => $action_plugin->getConfiguration()['weight']
        ];
      }
    }
    // These are core hooks
    $operations += \Drupal::moduleHandler()->invokeAll('entity_operation', [$transaction]);
    \Drupal::moduleHandler()->alter('entity_operation', $operations, $transaction);
    uasort($operations, '\Drupal\Component\Utility\SortArray::sortByWeightElement');
    return $operations;
  }


  /**
   * Utility function.
   *
   * Loads any of the transaction operation actions.
   *
   * @param string $operation
   *
   * @return Action
   */
  public static function loadOperation(string $operation) : Action {
    $name = 'transaction_' . $operation;
    if ($action = Action::load($name)) {
      return $action;
    }
    throw new \Exception('Action does not exist: '.$name);
  }

  /**
   * Utility function.
   * Load those special transaction operations which are also actions.
   *
   * @param bool $full
   *   TRUE to load the action objects
   *
   * @return []
   */
  public static function loadAllActions(bool $full = FALSE) : array {
    $actions = [];
    $config_names = \Drupal::configFactory()->listAll('system.action.transaction_');
    foreach ($config_names as $action_name) {
      $actions[] = substr($action_name, 26);
    }
    if ($full) {
      foreach ($config_names as $action_name) {
        $action_names[] = substr($action_name, 14);
      }
      $actions = array_combine($actions, Action::loadMultiple($action_names));
    }
    return $actions;
  }
}
